from datetime import datetime
from random import randint
from time import sleep

from beartype import beartype
from beartype.typing import Callable, Union
from google.protobuf.timestamp_pb2 import Timestamp


def get_datetime_from_pb_ts(timestamp):
    return datetime.fromtimestamp(timestamp.seconds + timestamp.nanos / 1e9)


@beartype
def get_pb_ts_from_datetime(python_timestamp: datetime) -> Timestamp:
    t = python_timestamp.timestamp()
    return Timestamp(seconds=int(t), nanos=int(t % 1 * 1e9))


@beartype
def random_sleep(
    min_sleep_seconds: int,
    max_sleep_seconds: int,
    pre_callback: Union[Callable, None] = None,
) -> None:
    random_sleep_seconds = randint(min_sleep_seconds, max_sleep_seconds)
    if pre_callback != None:
        pre_callback(random_sleep_seconds)
    sleep(randint(min_sleep_seconds, max_sleep_seconds))



from contextlib import asynccontextmanager

import asyncpg
import grpc
from loguru import logger


@asynccontextmanager
async def ErrorAsyncContextManager(context: any) -> asyncpg.Connection:
    try:
        yield None
    except Exception as e:
        message = type(e).__name__ + ": " + str(e)
        logger.error(message)
        await context.abort(grpc.StatusCode.UNKNOWN, message)



import os

import asyncpg
from beartype.typing import Any, Dict, List, Tuple, Union
from loguru import logger


# Serial Dictionary.
class Serial(dict):
    def __getitem__(self, key):
        return f"${list(self.keys()).index(key) + 1}"


# Pyformat to psql format.
def pyformat2psql(query: str, args_dict: Dict[str, Any]) -> Tuple[str, List[Any]]:
    # Remove args not present in query.
    args_list = list(args_dict.keys())
    for value in args_list:
        if f"{{{value}}}" not in query:
            args_dict.pop(value, None)
    # Generate query with serial positions.
    args = Serial(args_dict)
    query_formatted = query.format_map(args)
    args_formatted = list(args.values())
    return query_formatted, args_formatted


def get_database_credentials():
    env_var_mapping = {
        "DATABASE_PASSWORD": "password",
        "DATABASE_USERNAME": "username",
        "DATABASE_HOST": "timescaledb.timescaledb.svc.cluster.local",
        "DATABASE_PORT": 5432,
        "DATABASE_NAME": "",
    }

    credentials = {}

    for key, default in env_var_mapping.items():
        credentials[key] = os.getenv(key, default)

        if len(str(credentials[key])) == 0:
            raise Exception("The environment variable {} is empty.".format(key))

    # Build the database url and append
    credentials["DATABASE_URL"] = "postgresql+psycopg2://{}:{}@{}:{}/{}".format(
        credentials["DATABASE_USERNAME"],
        credentials["DATABASE_PASSWORD"],
        credentials["DATABASE_HOST"],
        credentials["DATABASE_PORT"],
        credentials["DATABASE_NAME"],
    )

    return credentials


async def get_database_pool() -> Union[asyncpg.pool.Pool, None]:
    # Access the database credentials from environment variables
    credentials = get_database_credentials()

    # Print debug information to the console
    logger.debug("Trying to connect to the database...")
    logger.debug("Host: {}".format(credentials["DATABASE_HOST"]))
    logger.debug("Port: {}".format(credentials["DATABASE_PORT"]))
    logger.debug("Database Name: {}".format(credentials["DATABASE_NAME"]))
    logger.debug("Username: {}".format(credentials["DATABASE_USERNAME"]))

    # Setup the database connection pool
    db_pool = await asyncpg.create_pool(
        user=credentials["DATABASE_USERNAME"],
        password=credentials["DATABASE_PASSWORD"],
        host=credentials["DATABASE_HOST"],
        port=credentials["DATABASE_PORT"],
        database=credentials["DATABASE_NAME"],
        statement_cache_size=0,
    )

    logger.debug("Successfully connected to the database.")

    return db_pool



from fsai_shared_funcs.proto_helpers.json_format import MessageToDict


# Move to fsai_shared_funcs
class ProtoToDict(object):
    def __init__(
        self,
    ):
        pass

    def __call__(self, f):
        async def wrapped_function(slf, request, context):
            request_as_dict = MessageToDict(
                request,
                including_default_value_fields=True,
                preserving_proto_field_name=True,
                float_precision=14,
            )

            result = f(slf, request_as_dict, context)  # Invoke the decorated function

            return await result

        return wrapped_function


# Move to fsai_shared_funcs
class ProtoToDictAsyncGenerator(object):
    def __init__(
        self,
    ):
        pass

    def __call__(self, f):
        async def wrapped_function(slf, request, context):
            request_as_dict = MessageToDict(
                request,
                including_default_value_fields=True,
                preserving_proto_field_name=True,
                float_precision=14,
            )

            # return await f(slf, request_as_dict, context)
            result = f(slf, request_as_dict, context)  # Invoke the decorated function

            if hasattr(
                result, "__aiter__"
            ):  # Check if result is an async iterator (yields)
                async for item in result:
                    yield item
            else:  # Result is not an async iterator, assumes it returns a value
                raise Exception(
                    "ProtoToDictAsyncGenerator should only be used on functions that return an async iterator"
                )

        return wrapped_function



from grpc_reflection.v1alpha import reflection
from loguru import logger

# Example service_configs
# service_configs = {
#     "AreaOfInterestApi": {
#         "pb_definition": area_of_interest_api_pb2,
#         "grpc_definition": area_of_interest_api_pb2_grpc,
#         "servicer_instance": AreaOfInterestApi(db_pool),
#     }
# }


class ServiceConfig:
    def __init__(self, server, service_configs):
        self.server = server
        self.service_configs = service_configs

        # Add each servicer to the server
        for service_name, config in self.service_configs.items():
            grpc_definition = config["grpc_definition"]
            servicer_instance = config["servicer_instance"]

            # Get the appropriate add_*Servicer_to_server function
            add_servicer_func = getattr(
                grpc_definition, f"add_{service_name}Servicer_to_server"
            )

            # Add the servicer instance to the server
            add_servicer_func(servicer_instance, self.server)

            logger.info(f"Added {service_name} to server")

    def enable_reflection(self):
        # For each pb_definition in service_configs, add it to the server reflection
        SERVICE_NAMES = tuple(
            [
                self.service_configs[service]["pb_definition"]
                .DESCRIPTOR.services_by_name[service]
                .full_name
                for service in self.service_configs
            ]
        )

        reflection.enable_server_reflection(SERVICE_NAMES, self.server)

        logger.info("Enabled server reflection")



from typing import Generator

from beartype import beartype
from loguru import logger


class DiskStorageAdapter:
    def __init__(self, base_path="/mnt/data/", chunk_size=1024):
        self.base_path = base_path
        self.chunk_size = chunk_size

    @beartype
    def get_bytes(self, file_key: str) -> bytes:
        # Get the bytes generator
        bytes_generator = self.get_bytes_generator(file_key)

        # Join the bytes
        bytes = b"".join(bytes_generator)

        # Return the bytes
        return bytes

    def get_bytes_generator(self, file_key: str) -> Generator[bytes, None, None]:
        full_key = "{}{}".format(self.base_path, file_key)

        logger.debug("Opening File: {}".format(full_key))

        try:
            with open(full_key, "rb") as f:
                while True:
                    bytes = f.read(self.chunk_size)

                    if not bytes:
                        break

                    yield bytes
        except Exception as e:
            logger.error("Error opening the file: {}".format(full_key))
            logger.error(e)
            raise e


# import os
# from beartype import beartype

# class S3StorageAdapter:
#     def __init__(self, s3_client, s3_bucket_name, chunk_size=1024):
#         self.s3_client = s3_client
#         self.s3_bucket_name = s3_bucket_name
#         self.chunk_size = chunk_size

#     def get_image_bytes(self, image_key):
#         # Access the image from storage via image key
#         obj = self.s3_client.get_object(
#             Bucket=self.s3_bucket_name,
#             Key=image_key,
#         )

#         # Access the file via chunked iterator
#         iterator = obj["Body"].iter_chunks(chunk_size=self.chunk_size)

#         while True:
#             try:
#                 bytes = next(iterator)

#                 if not bytes:
#                     break

#                 yield bytes

#             except Exception as e:
#                 break


# @beartype
# def get_storage_configuration() -> dict:
#     config_mapping = {
#         "S3_ACCESS_KEY": "",
#         "S3_SECRET_KEY": "",
#         "S3_BUCKET_NAME": "",
#     }

#     config = {}

#     for key, default in config_mapping.items():
#         config[key] = os.getenv(key, default)

#         if len(str(config[key])) == 0:
#             raise Exception("The environment variable {} is empty.".format(key))

#     return config



import functools

from loguru import logger


class RedisCacheManager:
    def __init__(self, _redis, redis_key: str):
        self._redis = _redis
        self.redis_key = redis_key

    def exists(self):
        logger.debug(f'Checking if key "{self.redis_key}" exists')
        return self._redis.exists(self.redis_key)

    def get(self):
        logger.debug(f'Getting key "{self.redis_key}"')
        return self._redis.pget(self.redis_key)

    def set(self, data: int, exp: int = 43200):
        logger.debug(f'Setting key "{self.redis_key}"')
        self._redis.pset(self.redis_key, data, ex=exp)


def cache_return_data(redis, key, exp=43200):
    def decorator_repeat(data_fn):
        @functools.wraps(data_fn)
        async def wrapper_repeat(*args, **kwargs):
            logger.debug(f'Caching return data for key "{key}"')
            # Create the RedisCacheManager
            r = RedisCacheManager(redis, key)

            # If the key doesn't exist
            if not r.exists():
                # Get the data from the function
                data = await data_fn()
                # Set the data in the cache
                r.set(data, exp)

            # Return the data
            return r.get()

        return wrapper_repeat

    return decorator_repeat



from contextlib import asynccontextmanager

import asyncpg
from loguru import logger


@asynccontextmanager
async def DatabaseAsyncContextManager(db_pool: asyncpg.pool.Pool) -> asyncpg.Connection:
    # Perform setup that would go in __aenter__

    logger.debug("Acquiring a connection from the db_pool")
    async with db_pool.acquire() as connection:
        async with connection.transaction():
            logger.debug("Acquired a connection from the db_pool")
            yield connection  # obj should be the object that will be bound in the as clause

    # Perform teardown that would go in __aexit__



from datetime import datetime

from ..time_helper import get_pb_ts_from_datetime


# TODO: Move this to fsai_shared_funcs
def dict_to_proto(dictionary, protobuf_message):
    # Iterate over the key-value pairs in the dictionary
    for key, value in dictionary.items():
        if isinstance(value, datetime):
            # Instead of setattr we need to use CopyFrom for composite fields
            # Otherwise we will get AttributeError:
            #   Assignment not allowed to composite field “field name” in protocol message object
            getattr(protobuf_message, key).CopyFrom(get_pb_ts_from_datetime(value))
        else:
            # Set the field value in the Protobuf message
            setattr(protobuf_message, key, value)

    return protobuf_message
